/*
    File:       ArgParse.h
 
    Function:	Argument parsing package
 
    Author:     Paul Heckbert
 
    Copyright:  
*/

#ifndef __ArgParse__
#define __ArgParse__

#include <stdio.h>

struct ArgForm
{
    /* a "form" contains the format, doc string, and miscellaneous internal */
    /* info about an argument.  It's an argument descriptor, basically */

    ArgForm *next;  	/* next in linked list */
    char *format;  		/* scanf-style format: "-size %d %F" */
    char *flag;  		/* flag portion of format:"-size" */
    char *code;  		/* just the format codes: "dF" */
    char *doc;  		/* documentation string: "set widget size" */
    short type;  		/* REGULAR | SIMPFLAG | PARAMFLAG |
    									   SUBRFLAG | SUBLISTFLAG | NOP */
    short nparam;  		/* number of parameters to flag */
    int parammask;  	/* bit i says ok to stop before param i, i=0..*/
    int **param;  		/* parameter pointer list */
    int (*subr)(int argc, char *argv[]);  	/* subroutine to call for action (if any) */
    ArgForm	*sublist;  	/* subordinate list (if any) */
    short rep;  		/* # times this flag repeated in arglist */
};

typedef ArgForm Arg_form;

/* form type values */
enum ArgTypes
{
	ARG_REGULAR =        1,	/* a regular argument */
	ARG_SIMPFLAG =       2,	/* a simple flag (no parameters) */
	ARG_PARAMFLAG =      3,	/* a flag with parameters */
	ARG_SUBRFLAG =       4,	/* a flag with subroutine action */
	ARG_SUBLISTFLAG =    5,	/* a sub-formlist */
	ARG_NOP =            6,	/* no arg or flag, just a doc string */
};

/* the following must be impossible pointer values (note: machine-dependent) */
enum ArgNextTypes
{
	ARG_MASKNEXT =	0x80000000,	/* mask for these NEXT flags */
	ARG_FLAGNEXT =	0x80000001,
	ARG_SUBRNEXT =	0x80000002,
	ARG_LISTNEXT =	0x80000003,
};

/* varargs tricks */
#define ARG_FLAG(ptr)		ARG_FLAGNEXT, (ptr)	/* for SIMPFLAG */
#define ARG_SUBR(ptr)		ARG_SUBRNEXT, (ptr)	/* for SUBRFLAG */
#define ARG_SUBLIST(ptr)	ARG_LISTNEXT, (ptr)	/* for SUBLISTFLAG */

/* error codes: BADCALL is a programmer error, the others are user errors */
enum ArgError
{
	ARG_BADCALL = 	-1,	/* arg_parse call itself is bad */
	ARG_BADARG = 	-2,	/* bad argument given */
	ARG_MISSING =	-3,	/* argument or parameter missing */
	ARG_EXTRA = 	-4,	/* extra argument given */
};

#define ARG_NARGMAX 10000	/* max number of allowed args */

extern int arg_debug, arg_doccol;
extern int arg_warning;  		/* print warnings about repeated flags? */

int 		arg_parse(int ac, char **av, ...);
int 		arg_parse_argv(int ac, char **av, ArgForm *form);
int 		arg_parse_stream(FILE *fp, ArgForm *form);
ArgForm 	*arg_to_form(int fish, ...);
			/* due to limitations of stdarg.h, you must pass in the
			 * number of fish you own as the first argument to
			 * arg_to_form */
void 		arg_form_print(ArgForm *form);
void		arg_form_append(ArgForm *form, ArgForm *additionalForm);

#endif
